
# coding: utf-8
#!/home/thomas/anaconda/bin/python

import os
import time
import datetime as dt
from collections import defaultdict
import random
import pickle
import pandas as pd
import numpy as np
import tensorflow as tf
import keras
from keras.models import load_model
import keras.backend as K
#from keras.utils import plot_model
import seaborn as sns
import feather
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV, PredefinedSplit
import matplotlib.pyplot as plt
from keras.utils.vis_utils import plot_model
import warnings
from sklearn.exceptions import UndefinedMetricWarning
import gc
from tsgen import TSGen
import mlflow
import mlflow.keras
import mlflow.sklearn
from keras_custom import f1, f1_loss, auc, get_callbacks, get_hmvts_model, get_cnn_model, get_lstm_model, get_mlp_model, save_history_plot, rebalance_dfs
from metrics import find_best_f1, find_best_fbeta, save_roc, get_result_df, f1_actual
import uuid
import os
import datetime as dt
import sys
import json
from sklearn.utils import class_weight
from sklearn.metrics import make_scorer

warnings.filterwarnings('ignore', category=UndefinedMetricWarning)
warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=FutureWarning)

EXP_PREFIX = "FINAL_EXPERIMENTS_"
PATH = "data/"
seed = 42
EARLY_STOPPING_PATIENCE = 3
SAVE_MODELS = True
RUN_RF = True
RUN_NN = True
CLASS_WEIGHT_FACTOR = "balanced_original"
VAL_FRAC = 0.15
TEST_FRAC = 0.15
SAMPLE = 1
EPOCHS = 20
LR = 1e-4
DROPOUT_FRAC = 0.5
W_REG = 0.001
A_REG = 0.001
batch_size = 512
fault_code = {"M104 Yaw System":1, "M099 TS-Converter System": 2, "M079 Control And Protection System":3, "M076 Blade Adjustment System":4, "M005 Out of Electrical Specification":5}
model_func = {"hmvts": get_hmvts_model, "lstm": get_lstm_model, "mlp": get_mlp_model, "cnn": get_cnn_model}

print("Reading input data...")
df = pd.read_parquet("data/df_sensors_labels.parquet.gzip")
label_cols = [c for c in df.columns if "Label" in c]

df["String"] = df["Turbine"].apply(lambda x: x[0])
df["Turbine_no"] = df["Turbine"].apply(lambda x: int(x[1:]))

df["Turbine"], turbine_map = pd.factorize(df["Turbine"])
df["String"], string_map = pd.factorize(df["String"])

df.loc[:,label_cols] = df.loc[:, label_cols].fillna(0)
df = df.fillna(-1)

first = True

for label_col in label_cols:
    seq_len = int(label_col.split("_")[1])
    if (seq_len not in [36,144]) or (label_col in ["Label_36_M104", "Label_36_M099"]): 
        continue #label_col = sys.argv[1] #'Label_36_M099'
    if not first:
        print("Reading new input data...")
        df = pd.read_parquet("data/df_sensors_labels.parquet.gzip")
        df["String"] = df["Turbine"].apply(lambda x: x[0])
        df["Turbine_no"] = df["Turbine"].apply(lambda x: int(x[1:]))

        df["Turbine"], turbine_map = pd.factorize(df["Turbine"])
        df["String"], string_map = pd.factorize(df["String"])

        df.loc[:,label_cols] = df.loc[:, label_cols].fillna(0)
        df = df.fillna(-1)
    first = False
    print("original shape", df.shape)
    #df = df.loc[df["TimeStamp"]<dt.datetime(2018,9,30)]
    #print("new shape after cutting 6 months", df.shape)
    if SAMPLE<1:
        print("Downsampling df to "+str(SAMPLE*100)+ "%")
        #total_days = (df["TimeStamp"].max()-df["TimeStamp"].min()).days
        #print(total_days)
        #df = df.loc[df["TimeStamp"]<(df["TimeStamp"].min()+pd.Timedelta(int(total_days*SAMPLE), unit="d"))]
        #print("new shape", df.shape)
        #print(df["TimeStamp"].min())
        #print(df["TimeStamp"].max())
        df = df.sample(frac=SAMPLE, random_state=seed)

    EXPERIMENT_NAME = EXP_PREFIX+label_col
    print("Running experiment:")
    print(EXPERIMENT_NAME)
    RUN_ID = dt.datetime.now().strftime("%Y-%m-%d-%H-%M-%S") 
    SAVE_DIR = "save_dir/"+str(EXPERIMENT_NAME)+"/"+RUN_ID+"/"
    os.makedirs(SAVE_DIR)
    seq_len = int(label_col.split("_")[1])

    mlflow.set_experiment(EXPERIMENT_NAME)
    # ## Data import and cleaning
    with mlflow.start_run() as run:
        # Log params
        mlflow.log_param("SEED", seed)
        mlflow.log_param("LABEL_COL", label_col)
        mlflow.log_param("EARLY_STOPPING_PATIENCE", EARLY_STOPPING_PATIENCE)
        mlflow.log_param("SAVE_MODELS", SAVE_MODELS)
        mlflow.log_param("RUN_RF", RUN_RF)
        mlflow.log_param("RUN_NN", RUN_NN)
        mlflow.log_param("CLASS_WEIGHT_FACTOR", CLASS_WEIGHT_FACTOR)
        mlflow.log_param("VAL_FRAC", VAL_FRAC)
        mlflow.log_param("TEST_FRAC", TEST_FRAC)
        mlflow.log_param("SAMPLE", SAMPLE)
        mlflow.log_param("EPOCHS", EPOCHS)
        mlflow.log_param("LR", LR)
        mlflow.log_param("DROPOUT_FRAC", DROPOUT_FRAC)
        mlflow.log_param("W_REG", W_REG)
        mlflow.log_param("A_REG", A_REG)
        mlflow.log_param("BATCH_SIZE", batch_size)
        mlflow.log_param("SEQ_LEN", seq_len)

        f1_actual_scorer = make_scorer(f1_actual, needs_proba=True, win=seq_len)
        month_dist = df[label_col].groupby([df["TimeStamp"].dt.year, df["TimeStamp"].dt.month]).sum().plot(kind="bar").get_figure()
        month_dist.savefig(SAVE_DIR+"month_dist.png")
        

        turb_dist = df.groupby("Turbine")[label_col].sum().plot(kind="bar").get_figure()
        turb_dist.savefig(SAVE_DIR+"turb_dist.png")

        df.sort_values("TimeStamp", inplace=True)


        get_split_time = lambda pct, label: df.loc[df[label].cumsum()>df[label].sum()*pct]["TimeStamp"].values[0] #.idxmin()

        df.sort_values(["TimeStamp", "Turbine"], inplace=True)

        print("Splitting into train, validation and test set...")

        val_ts = get_split_time(1-VAL_FRAC-TEST_FRAC, label_col) # dt.datetime(2018,6,30) #
        test_ts = get_split_time(1-TEST_FRAC, label_col) # dt.datetime(2018,9,30)

        mlflow.log_param("Training_set", "from "+str(df["TimeStamp"].min())+"to "+str(val_ts))
        mlflow.log_param("Validation_set", "from "+str(val_ts)+" to "+str(test_ts))
        mlflow.log_param("Test_set", "from "+str(test_ts)+"to "+str(df["TimeStamp"].max()))

        df.sort_values(["Turbine","TimeStamp"], inplace=True)

        # NB: Switched here to test 
        df_train = df.loc[df["TimeStamp"]<val_ts]
        df_val = df.loc[(df["TimeStamp"]>=val_ts) & (df["TimeStamp"]<test_ts)]
        df_test = df.loc[df["TimeStamp"]>=test_ts]

        df_train, df_val, df_test = rebalance_dfs([df_train, df_val, df_test], label_col, seed)
        
        mlflow.log_param("Number of samples in Training set",df_train.shape[0])
        mlflow.log_param("Number of samples in Validation set",df_val.shape[0])
        mlflow.log_param("Number of samples in Test set",df_test.shape[0])

        mlflow.log_param("Number of PF-labels in Training set", df_train[label_col].sum())
        mlflow.log_param("Number of PF-labels in Validation set", df_val[label_col].sum())
        mlflow.log_param("Number of PF-labels in Test set", df_test[label_col].sum())

        mlflow.log_param("Fraction of PF-labels in Training set", df_train[label_col].sum()/df_train.shape[0])
        mlflow.log_param("Fraction of PF-labels in Validation set", df_val[label_col].sum()/df_val.shape[0])
        mlflow.log_param("Fraction of PF-labels in Test set", df_test[label_col].sum()/df_test.shape[0])

        X_train, y_train = df_train.drop(["TimeStamp", "Remove"]+label_cols, axis=1), df_train[label_col].values
        X_val, y_val = df_val.drop(["TimeStamp", "Remove"]+label_cols, axis=1), df_val[label_col].values
        X_test, y_test = df_test.drop(["TimeStamp", "Remove"]+label_cols, axis=1), df_test[label_col].values
        del df_train, df_val, df_test, df
        gc.collect()

        emb_cols = [i for i,c in enumerate(X_train.columns) if c in ["Turbine", "String"]]
        emb_col_card = {i:len(X_train.iloc[:,i].unique()) for i in emb_cols}
        print("Cardinality:")
        print(emb_col_card)
        num_cols = [i for i,c in enumerate(X_train.columns) if (c not in ["Turbine", "String"])] 

        #Compute class weights
        if CLASS_WEIGHT_FACTOR == "balanced":
            class_weights = class_weight.compute_class_weight("balanced", np.unique(y_train), y_train) #{0:len(y_train)/((y_train==0).sum()), 1:CLASS_WEIGHT_FACTOR*len(y_train)/((y_train==1).sum())}
            if type(class_weights)!=dict:
                new_cw = {c:w for c,w in enumerate(class_weights)}
                class_weights = new_cw
        elif CLASS_WEIGHT_FACTOR == "balanced_original":
            #Balance with original number of faults
            class_weights = {0:1, 1:len(y_train)/(pd.Series(y_train).diff()==-1).astype(int).sum()}
        elif type(CLASS_WEIGHT_FACTOR) == float:
            class_weights = {0:1, 1:CLASS_WEIGHT_FACTOR*(len(y_train)/((pd.Series(y_train)==1).astype(int).sum()/seq_len))}
        
        mlflow.log_param("class_weights", class_weights)
        # ## Grid search + modelling RF
        results = {}
        # In[40]:
        if RUN_RF:
            print("Working on Random forest preprocessing...")

            X_tv = np.concatenate([X_train.values, X_val.values])
            y_tv = np.concatenate([y_train, y_val])
            split_inds = np.empty_like(y_tv)
            split_inds[:] = -1
            split_inds[len(X_train):] = 0
            split = PredefinedSplit(split_inds)

            model_type = "rf"
            print("Working on "+model_type+" model")
            print("---------------------------------")
            results[model_type] = {}
            
            roc_plot_name = SAVE_DIR+model_type+"_roc_"
            param_grid = {'max_depth': [1, 3, 5, 7, 10], 'max_features': [5, 8, 10, 20]}
            grid = GridSearchCV(RandomForestClassifier(random_state=seed, class_weight=class_weights), cv=split, param_grid=param_grid, n_jobs=1, scoring=f1_actual_scorer)
            grid.fit(X_tv, y_tv)
            mlflow.log_metric(model_type+"_val_f1", grid.best_score_)

            preds_val = grid.predict_proba(X_val)
            roc_auc_val = save_roc(y_val, preds_val[:,1], roc_plot_name+"val.png")
            mlflow.log_artifact(roc_plot_name+"val.png")
            mlflow.log_metric(model_type+"_val_roc_auc", roc_auc_val)
            preds = grid.predict_proba(X_test)
            roc_auc = save_roc(y_test, preds[:,1], roc_plot_name+"test.png")
            mlflow.log_artifact(roc_plot_name+"test.png")
            mlflow.log_metric(model_type+"_test_roc_auc", roc_auc)
            result_f1 = find_best_f1(y_test, preds[:,1], seq_len)
            results[model_type]["f1"] = result_f1
            result_fb = find_best_fbeta(y_test, preds[:,1], seq_len, beta=4)
            results[model_type]["f_beta"] = result_fb
            for metric, value in result_f1.items():
                mlflow.log_metric(metric+"_"+model_type+"_f1", value)  
            results[model_type]["f_beta"] = result_fb
            for metric, value in result_fb.items():
                mlflow.log_metric(metric+"_"+model_type+"_f_beta", value)  
            print(results)
            df_cv = pd.DataFrame(grid.cv_results_)
            df_cv_name = SAVE_DIR+"df_cv.csv"
            df_cv.to_csv(df_cv_name)
            mlflow.log_artifact(df_cv_name)
            del X_tv, y_tv
            gc.collect()

        if RUN_NN:
            print("Working on NN preprocessing...")
            scaler = StandardScaler()
            X_train.iloc[:,num_cols] = scaler.fit_transform(X_train.iloc[:, num_cols].values)
            X_val.iloc[:,num_cols] = scaler.transform(X_val.iloc[:, num_cols].values)
            X_test.iloc[:,num_cols] = scaler.transform(X_test.iloc[:, num_cols].values)

            train_gen = TSGen(X_train,y_train, length=seq_len, batch_size=batch_size, shuffle=True, emb_cols=emb_cols, num_cols=num_cols, seed=seed)
            train_gen_det = TSGen(X_train,y_train, length=seq_len, batch_size=batch_size, shuffle=False, emb_cols=emb_cols, num_cols=num_cols, seed=seed)
            val_gen = TSGen(X_val, y_val, length=seq_len, batch_size=batch_size, emb_cols=emb_cols, num_cols=num_cols, seed=seed)
            test_gen = TSGen(X_test, y_test, length=seq_len, batch_size=batch_size, emb_cols=emb_cols, num_cols=num_cols)

            ytr = [train_gen_det[i][1] for i in range(len(train_gen_det))]
            y_tr = np.concatenate(ytr)
            print("y_train shape", y_tr.shape)

            yv = [val_gen[i][1] for i in range(len(val_gen))]
            y_v = np.concatenate(yv)
            print("y_val shape", y_v.shape)

            yt = [test_gen[i][1] for i in range(len(test_gen))]
            y_t = np.concatenate(yt)
            print("y_test shape", y_t.shape)

            del ytr, yv, yt

        for model_type in model_func.keys():
            if not RUN_NN: continue
            print("Working on "+model_type+" model")
            print("---------------------------------")
            #if (model_type == "hmvts"): #in ["mlp", "tcn"]: 
            #    continue
            results[model_type] = {}
            if model_type == "hmvts":
                #DROPOUT_FRAC = 0.8
                W_REG = W_REG*5
                A_REG = A_REG*5
            model = model_func[model_type](emb_col_card, num_cols, seq_len, LR, DROPOUT_FRAC, W_REG, A_REG)
            model_arch_name = SAVE_DIR+model_type+".png"
            hist_plot_name = SAVE_DIR+"training_history_"+model_type+".png"
            model_sum_name = SAVE_DIR+model_type+".txt"
            model_result_name = SAVE_DIR+"final_results.txt"
            print(model.summary())
            with open(model_sum_name,'w') as f:
                model.summary(print_fn=lambda x: f.write(x+"\n"))
            mlflow.log_artifact(model_sum_name)
            plot_model(model=model, to_file=model_arch_name)
            mlflow.log_artifact(model_arch_name)
            model_save_dir = SAVE_DIR+model_type+'_best.h5'
            callbacks = get_callbacks(model_save_dir, EARLY_STOPPING_PATIENCE, train_gen_det, y_tr, val_gen, y_v, seq_len, SAVE_MODELS)
            hist = model.fit_generator(generator=train_gen,class_weight=class_weights, callbacks=callbacks, shuffle=True, validation_data=val_gen, epochs=EPOCHS, verbose=1)
            model = load_model(model_save_dir, custom_objects={"f1_loss":f1_loss, "f1":f1})
            save_history_plot(hist, hist_plot_name)
            mlflow.log_artifact(hist_plot_name)
            set_dict =  {"train":(train_gen_det,y_tr), "val":(val_gen, y_v), "test":(test_gen,y_t)}
            for pred_name, (gen,y_set) in set_dict.items():
                preds = model.predict_generator(gen)
                if pred_name=="test":
                    t_preds = preds
                print("Preds shape:", preds.shape)
                preds = preds.flatten()
                roc_plot_name = SAVE_DIR+model_type+"_roc_"+pred_name+".png"
                roc_auc = save_roc(y_set, preds, roc_plot_name)
                mlflow.log_metric(model_type+"_pred_"+pred_name+"_auc", roc_auc)
                mlflow.log_artifact(roc_plot_name)
            result_f1 = find_best_f1(y_t, t_preds, seq_len)
            results[model_type]["f1"] = result_f1
            for metric, value in result_f1.items():
                mlflow.log_metric(metric+"_"+model_type+"_f1", value)  
            result_fb = find_best_fbeta(y_t, t_preds, seq_len, beta=4)
            results[model_type]["f_beta"] = result_fb
            for metric, value in result_fb.items():
                mlflow.log_metric(metric+"_"+model_type+"_f_beta", value)  
            print(results)
            with open(model_result_name,'w') as f:
                f.write(json.dumps(str(results)))
            mlflow.log_artifact(model_result_name)
        res_df = get_result_df(results)
        #save df
        df_res_name = SAVE_DIR+"result_df.csv"
        res_df.to_csv(df_res_name)
        mlflow.log_artifact(df_res_name)
